/*#############################################################################
# IPXSPX.CPP
#
# 19.05.1997 (Codor)
#
# Reference:
# - Network Programming in C by Barry Nance (1990, Que corporation)
# - NetWare System Interace Technical Overview
# - Netware C Interface - DOS
# - NetWare Programmer's Guide by Ralph Davis
# Intenet pages:
# - http://occam.sjf.novell.com:8080/nwsdk.redsdk08.english/a_transp/303.toc#X
#   (Contains complete reference to IPX functions)
#############################################################################*/

#include <dos.h>
#include <mem.h>
#include <stdlib.h>
#include "types.h"
#include "ipxspx.h"

static TIpxSpxEntryPoint IpxSpxEntry = NULL;

//=============================================================================
// struct TIpxAddress
//
//=============================================================================

TIpxAddress::TIpxAddress(bool isbroadcast)
{
  memset(this,'\0',sizeof(*this));
  if (isbroadcast)
    memset(this->Node,0xFF,6);
}

//
// Local address is the address of the local machine Ethernet's card
//
void TIpxAddress::GetLocalAddress(void)
{
  TNetworkAddress na;
  IPXGetInternetworkAddress(&na);
  memcpy(Network,na.NetworkNumber,4);
  memcpy(Node,na.NodeAddress,6);
}

bool TIpxAddress::IsBroadcast(void)
{
  char cmpstr[6];
  memset(&(cmpstr[0]),0xFF,6);
  return (memcmp(this->Node,&(cmpstr[0]),6) == 0) ? True : False;
}

bool TIpxAddress::IsSame(TIpxAddress *addr, bool useSocket)
{
  if (memcmp(&(this->Node[0]),&(addr->Node[0]),6) == 0) {
    if (useSocket && (memcmp(&(this->Socket[0]),&(addr->Socket[0]),2) != 0))
      return False;
    return True;
  }
  return False;
}

//=============================================================================
// struct TIpxHeader
//
//=============================================================================

TIpxHeader::TIpxHeader()
{
  memset(this,'\0',sizeof(*this));
  PacketType = 4;
}

void TIpxHeader::SetDestAddress(TIpxAddress *dest)
{
  memcpy(&Dest,dest,sizeof(TIpxAddress));

}

void TIpxHeader::SetSrcAddress(TIpxAddress *src)
{
  memcpy(&Src,src,sizeof(TIpxAddress));
}

//=============================================================================
// struct TEventControlBlock
//
//=============================================================================

TEventControlBlock::TEventControlBlock()
{
  memset(this,'\0',sizeof(*this));
  FragmentCount = 2;
}

//=============================================================================
// class TIpxSocket
// This class receives (via constructor) a socket number in low-high order. It
// is internally converted to high-low and kept in this format. If requested
// (via GetSocket()), return value is high-low
//=============================================================================

//
// Input value is low-high
//
TIpxSocket::TIpxSocket(word socket, bool create, word longevity)
{
  Longevity = longevity;
  Socket = Swap(socket);
  ErrorCode = 0;
  ExErrorCode = TIpxStatus::IpxSuccess;
  Created = create;
}

TIpxSocket::~TIpxSocket()
{
  Destroy();
}

//
// Open a socket. Socket is not opened in constructor since this will prevent
// static object of TIpxSocket 'cause IPX is not yet initialised
//
word TIpxSocket::Create(void)
{
  if (Created) {
    ErrorCode = IPXOpenSocket(Longevity,&Socket);
    switch (ErrorCode) {
      case 0x00: ExErrorCode = TIpxStatus::IpxSuccess; break;
      case 0xFE: ExErrorCode = TIpxStatus::IpxSocketTableIsFull; break;
      case 0xFF: ExErrorCode = TIpxStatus::IpxSocketAlreadyOpen; break;
    }
  }
  return ExErrorCode;
}

//
// Close a socket. This is implemented as a separate structure since the
// object can be created as a static one and there are problems with
// destructors (which one is executed first).
//
void TIpxSocket::Destroy(void)
{
  if (Created)
    IPXCloseSocket(Socket);
}

//
// Returned value is high-low
//
word TIpxSocket::GetSocket(void)
{
  return Socket;
}

//
// Returned value is low-high
//
word TIpxSocket::GetLowHighSocket(void)
{
  return Swap(Socket);
}

word TIpxSocket::GetErrorCode(void)
{
  return ExErrorCode;
}

//=============================================================================
// class TIpx
//
//=============================================================================

// Static function
bool TIpx::Init(void)
{
  return IPXInitialize();
}

TIpx::TIpx(TIpxSocket *socket, TEventServiceRoutine proc, word /*dataSize*/)
{
  ErrorCode = 0;
  ExErrorCode = TIpxStatus::IpxSuccess;
  DataLength = sizeof(Data);
  memset(&(Data[0]),'\0',DataLength);
  SocketPtr = socket;
  Ecb.SocketNumber = SocketPtr->GetSocket();
  Ecb.EventServiceRoutine = proc;
}

TIpx::~TIpx()
{
  Cancel();
  while (!IsCompleted())
    ;
}

//
// If the object is created with a default constructor (Socket = NULL), I
// have to assign a socket to this object after construction. This is a
// socket on which the object is listening to.
//
void TIpx::SetSocket(TIpxSocket *socket)
{
  SocketPtr = socket;
  Ecb.SocketNumber = SocketPtr->GetSocket();
}

//
// Destination socket does not have to be opened
//
void TIpx::SetDestination(TIpxAddress *dest, TIpxSocket *socket)
{
  word sock = socket->GetSocket();    // sock is low-high
  memcpy(&(dest->Socket[0]),&sock,sizeof(word));
  Header.SetDestAddress(dest);
  /*** No need for this since we're always on the same network
  TNodeAddress na;
  ErrorCode = IPXGetLocalTarget(dest,&na);
  switch (ErrorCode) {
    case 0x00: ExErrorCode = TIpxStatus::IpxSuccess; break;
    case 0xFA: ExErrorCode = TIpxStatus::IpxNoPathToDestination; break;
  };
  memcpy(&(Ecb.ImmediateAddress[0]),&(na.NodeAddress[0]),6);
  ***/
  memcpy(&(Ecb.ImmediateAddress[0]),&(Header.Dest.Node[0]),6);
}

//
// data length must not be greater then 546 bytes
//
int TIpx::Send(void *data, int length)
{
  if (Ecb.InUse != TIpxStatus::EcbInUseFree)
    return TIpxStatus::IpxAlreadyInUse;
  if (length > 546)
    return TIpxStatus::IpxDataTooLong;
  memset(Data,'\0',sizeof(Data)/sizeof(Data[0]));
  DataLength = length;
  if (data)
    memcpy(Data,data,DataLength);
  Header.Checksum = 0;
  Header.Length = 0;
  memset(&(Header.Src),'\0',sizeof(TIpxAddress));
  //Header.Src.GetLocalAddress();
  Ecb.FragmentCount = 2;
  Ecb.Fragment[0].Address = &Header;
  Ecb.Fragment[0].Length = sizeof(TIpxHeader);
  Ecb.Fragment[1].Address = &(Data[0]);
  Ecb.Fragment[1].Length = DataLength;
  IPXSendPacket(&Ecb);
  return TIpxStatus::IpxSuccess;
}

int TIpx::Listen(void)
{
  if (Ecb.InUse != TIpxStatus::EcbInUseFree)
    return TIpxStatus::IpxAlreadyInUse;
  DataLength = sizeof(Data);
  Ecb.FragmentCount = 2;
  Ecb.Fragment[0].Address = &Header;
  Ecb.Fragment[0].Length = sizeof(TIpxHeader);
  Ecb.Fragment[1].Address = &(Data[0]);
  Ecb.Fragment[1].Length = DataLength;
  ErrorCode = IPXListenForPacket(&Ecb);
  switch (ErrorCode) {
    case 0x00: ExErrorCode = TIpxStatus::IpxSuccess; break;
    case 0xFF: ExErrorCode = TIpxStatus::IpxSocketNotOpen; break;
  };
  return ExErrorCode;
}

int TIpx::Cancel(void)
{
  ErrorCode = IPXCancelEvent(&Ecb);
  switch (ErrorCode) {
    case 0x00: ExErrorCode = TIpxStatus::IpxSuccess; break;
    case 0xF9: ExErrorCode = TIpxStatus::IpxCannotCancelEcb; break;
    case 0xFF: ExErrorCode = TIpxStatus::IpxEcbIsNotInUse; break;
  };
  return ExErrorCode;
}

//
// Length of received packet's data part is in the Length field of the Ipx
// header (must substract size of header)
//
int TIpx::GetData(void far *data, int length)
{
  int len = Swap(Header.Length);
  len -= sizeof(TIpxHeader);
  if (len > length)
    len = length;
  memcpy(data,&(Data[0]),len);
  return len;
}

TIpxHeader far *TIpx::GetHeader(TIpxHeader *hdr) const
{
  if (hdr)
    memcpy(hdr,&Header,sizeof(TIpxHeader));
  return (TIpxHeader far *)&Header;
}

TEventControlBlock far *TIpx::GetEcb(TEventControlBlock *ecb) const
{
  if (ecb)
    memcpy(ecb,&Ecb,sizeof(TEventControlBlock));
  return (TEventControlBlock far *)&Ecb;
}

word TIpx::GetErrorCode(void)
{
  return ExErrorCode;
}

//
// Let Ipx work and then check if it's Ecb is free
//
bool TIpx::IsCompleted(void)
{
  IPXRelinquishControl();
  return (Ecb.InUse) ? False : True;
}

bool TIpx::IsFree(void)
{
  return (Ecb.InUse == 0) ? True : False;
}

//=============================================================================
// IPX functions
// Each function returns an error code 0x0x01 if the IPX is not initialized
// (function IPXInitialize()) which means that the static pointer IpxSpxEntry
// is not valid (is NULL).
//=============================================================================

//
// Must be called before any other function since this function will return
// a pointer to IPX/SPX entry point.
// The IPX major and minor version values are stored in two bytes pointed to
// by ES:BX after calling Int 2Fh AX=7A00h which gets the IPX far call handler
// address.
//
bool IPXInitialize(void)
{
  /*
  union REGS regs;
  struct SREGS sregs;
  regs.x.ax = 0x7A00;
  int86x(0x2F,&regs,&regs,&sregs);
  if (regs.h.al != 0xFF)
    return False;
  IpxSpxEntry = (TIpxSpxEntryPoint)MK_FP(sregs.es,regs.x.di);
  */
  _AX = 0x7A00;
  asm int 02Fh;
  if (_AL != 0xFF)
    return False;
  IpxSpxEntry = (TIpxSpxEntryPoint)MK_FP(_ES,_DI);
  return True;
}

//
// Will not execute ESR! Will set InUse flag to != 0. After canceling, InUse
// will be set to 0.
// Return code:
//  0x00 - success
//  0xF9 - ECB cannot be canceled
//  0xFF - ECB is not in use
// Possible ECB completion code:
//  0xFC - Event canceled
//
word IPXCancelEvent(TEventControlBlock far *ecb)
{
  if (IpxSpxEntry) {
    _BX = 0x0006;
    _ES = FP_SEG(ecb);
    _SI = FP_OFF(ecb);
    IpxSpxEntry();
    return _AL;
  }
  return 0x01;
}

//
// Sockets must be closed before the program terminates. Closing a socket will
// cancel any events associated with a socket.
// Will returns a value of FCh in ECB's Completion Code field indicating the
// event has been cancelled. Sets each ECB's In Use field to 00h.
// IPX does not call the ECBs' associated Event Service Routines.
// Any socket number can be closed; no error is generated if the specified
// socket is not open. Must not be called from within an Event Service Routine.
// socket is low-high order (this is from the Novell Web site)
//
void IPXCloseSocket(word socket)
{
  if (IpxSpxEntry) {
    _BX = 0x01;
    _DX = socket;
    IpxSpxEntry();
  }
}

//
// Courtesy to the other (listening) side. Must not be called from within
// an ESR. Must not be called from within an Event Service Routine.
// All fields in *rb are high-low.
//
void IPXDisconnectFromTarget(TIpxAddress far *rb)
{
  if (IpxSpxEntry) {
    _BX = 0x000B;
    _ES = FP_SEG(rb);
    _SI = FP_OFF(rb);
    IpxSpxEntry();
  }
}

//
// Returns information for the local station.
//
void IPXGetInternetworkAddress(TNetworkAddress far *rb)
{
  if (IpxSpxEntry) {
    _BX = 0x0009;
    _ES = FP_SEG(rb);
    _SI = FP_OFF(rb);
    IpxSpxEntry();
  }
}

//
// Use DOS timer tick instead of this.
// Returned value is in high-low order.
// The result is accurate even if the clock cycle wraps to zero between the
// first and second event. The result is not accurate if the clock cycle
// wraps to zero twice. This timer is not intended for use with large time
// intervals.
//
word IPXGetIntervalMarker(void)
{
  if (IpxSpxEntry) {
    _BX = 0x0008;
    IpxSpxEntry();
    return _AX;
    // ulong far *tickPtr = MK_FP(0x0040,0x006C);
    // ulong then = *tickPtr;
    // do something...
    // ulong now = *tickPtr;
  }
  return 0;
}

//
// For a given destination (rb), it returns a routing information (na). Copy
// this into the ECB ImmediateAddress field. (na) can be either bridge address
// or the destination station itself (no bridge found).
// When IPX receives a packet, it records the node address of the sending node
// in the Immediate Address field of the receive ECB. Therefore, once an
// application begins exchanging packets with an application on another node,
// the application can search the Immediate Address field of each receive ECB
// to obtain the Local Target value instead of continually calling
// IPXGetLocalTarget. IPX Get Local Target can be called from within an IPX
// Event Service Routine, an AES Event Service Routine, or directly from the
// main portion of an application. It cannot be called from any other kind of
// interrupt service routine.
// Return code:
//  0x00 - Success
//  0xFA - No path or destination
// *rb and *na are both high-low
//
word IPXGetLocalTarget(TIpxAddress far *rb, TNodeAddress far *na)
{
  if (IpxSpxEntry) {
    _BX = 0x0002;
    _ES = FP_SEG(rb);
    _SI = FP_OFF(rb);
    //_ES = FP_SEG(na);
    _DI = FP_OFF(na);
    IpxSpxEntry();
    // na = (TNodeAddress far *)MK_FP(_ES,_DI);

    // CX contains estimated transport time in timer ticks (18.2)
    // The Transport Time is an estimate of the time a 576-byte packet takes to
    // travel from the source node to the destination node (in clock ticks of
    // approximately 1/18 second). Since the returned value is only an estimate,
    // actual travel time may vary depending on network traffic and packet size.
    // AH contains number of hops to the destination station
    return _AL;
  }
  return 0x01;
}

//
// IPX returns immediately and waits for the packet in background.
// Following ECB fields must be set:
//  SocketNumber        - previously opened with IPXOpenSocket
//  EventServiceRoutine - pointer or NULL
//  FragmentCount       - typically a value of 2 (header and data)
//  Fragment[0].Address - far pointer to IPX header buffer
//  Fragment[0].Length  - length of the IPX header (30)
//  Fragment[1].Address - far pointer for message data area (max 546 bytes)
//  Fragment[1].Length  - length of data area
// Return code:
//  0x00 - success
//  0xFF - listening socket does not exists
// ECB completion code:
//  0x00 - packet received successfully
//  0xFC - this event is canceled
//  0xFD - buffer too small
//  0xFF - socket not open
//
word IPXListenForPacket(TEventControlBlock far * /*ecb*/)
{
  if (IpxSpxEntry) {
    /*
    _BX = 0x04;
    _ES = FP_SEG(ecb);
    _SI = FP_OFF(ecb);
    */
    asm {
      mov bx,0004h;
      mov si,[bp+6];
      mov es,[bp+8];
    }
    IpxSpxEntry();
    return _AL;
  }
  return 0x01;
}

//
// If the specified *socket is 0, then IPX will allocate a socket in the range
// 0f 0x4000 to 0x8000.
// *socket has a high-low organization.
// longevity can be 0x00 for short-lived or 0xFF for long lived
// Terminate-and-stay-resident applications that need to communicate via IPX
// packets should specify FFH to leave the socket open. All other applications
// should set this flag to 00h to ensure the socket is closed when the application
// (accidentally or purposefully) terminates.
// By default, IPX supports up to 20 open sockets on one workstation. The value
// can be configured to open up to 150 sockets on one workstation.
// Return code:
//  0x00 - success
//  0xFE - socket table is full
//  0xFF - Socket already open
//
word IPXOpenSocket(word longevity, word *socket)
{
  if (IpxSpxEntry) {
    _AL = longevity;
    _DX = *socket;      // this line scrambles _BX
    _BX = 0x0000;
    IpxSpxEntry();
    *socket = _DX;
    return _AL;
  }
  return 0x01;
}

//
// IPX Relinquish Control relinquishes control of a workstation CPU from
// the application, so that other processing can be handled while the
// application is waiting for input.
//
// IPX applications should make repeated calls to this function during
// idle time to allow other applications in the computer to use the CPU.
// This is especially important when the application is resident with
// NetWare server or bridge software because it greatly improves the
// efficiency of the server or bridge.
//
// A call to this function also informs the workstation communications
// driver that the CPU is temporarily free.  This is important if the
// driver is not interrupt driven.  On a normal workstation, this function
// invokes a polling procedure provided by the network communications
// driver.  It represents the only opportunity the driver has to use the
// CPU to send and receive packets; events essential to the application
// itself.
// es:si=ECB address (Is this OK?)
//
void IPXRelinquishControl(void)
{
  if (IpxSpxEntry) {
    _BX = 0x000A;
    IpxSpxEntry();
  }
}

//
// Will send a packet in the background.
// An application can send or broadcast an IPX packet to any socket on
// the internetwork including the socket on which the application sends
// the packet. Packets sent to sockets that reside in the same node as
// the application are called intranode packets.
// Following ECB fields must be set:
//  SocketNumber        - previously opened with IPXOpenSocket
//  EventServiceRoutine - function pointer or NULL
//  ImmediateAddress    - target node or bridge address
//  FragmentCount       - Equals 2 (for header and data)
//  Fragment[0]         - Defines area for header
//  Fragment[1]         - Defines area for data
// Following IPXHeader fields must be set:
//  PacketType          - Equals 4
//  DestNetworkNumber
//  DestNodeNumber
//  DestSocket
// ECB completion code:
//  0x00 - packet sent successfully
//  0xFC - this event is canceled
//  0xFD - bad packet
//  0xFE - undeliverable packet
//  0xFF - hardware or network failure
//
void IPXSendPacket(TEventControlBlock far * /*ecb*/)
{
  if (IpxSpxEntry) {
    /*
    _BX = 0x03;
    _ES = FP_SEG(ecb);
    _SI = FP_OFF(ecb);
    */
    asm {
      mov bx,0003h;
      mov si,[bp+6];
      mov es,[bp+8];
    }
    IpxSpxEntry();
  }
}

#if 0     // We are not using SPX here !!!

//
// Generates a checksum for a packet defined in ecb. Will be placed into the
// header's Checksum field. Do not modify packet after the checksum is
// generated.
// IPX Generate Checksum enables interrupts and is fully re-entrant.
// IPX Generate Checksum is provided for programs that wish to use
// IPXFastSend to send checksummed packets. As with IPXFastSend, the caller
// must initialize all IPX header fields. However, IPX Generate Checksum will
// initialize the Checksum and TransportControl fields.
//
void IPXGenerateChecksum(TEventControlBlock far *ecb)
{
  if (IpxSpxEntry) {
    _BX = 0x21;
    _ES = FP_SEG(ecb);
    _SI = FP_OFF(ecb);
    IpxSpxEntry();
  }
}

int IPXGetMaxPacketSize(void)
{
  if (IpxSpxEntry) {
    _BX = 0x001A;
    IpxSpxEntry();
    // _CX contains an IPX retry count
    // Max paxket size will never be less then 576
    return _AX;
  }
  return 0x01;
}

//
// IPX Send With Checksum behaves exactly the same as IPXSendPacket,
// except that a checksum is generated for the packet data and placed in
// the IPX checksum field before it is transmitted.
//
void IPXSendWithChecksum(TEventControlBlock far *ecb)
{
  if (IpxSpxEntry) {
    _BX = 0x0020;
    _ES = FP_SEG(ecb);
    _SI = FP_OFF(ecb);
    IpxSpxEntry();
  }
}

//
// IPX Verify Checksum determines whether or not the packet data received
// is corrupt. This is done by generating a new checksum for the data and
// comparing it to the checksum in the IPX header checksum field.
//
int IPXVerifyChecksum(TEventControlBlock far *ecb)
{
  if (IpxSpxEntry) {
    _BX = 0x0022;
    _ES = FP_SEG(ecb);
    _SI = FP_OFF(ecb);
    IpxSpxEntry();
    return _AX;
  }
  return 0x01;
}

#endif

//============================================================================
// SPX functions
//
//============================================================================

#if 0

void SPXAbortConnection(word connectionID)
{
  _BX = 0x14;
  _DX = connectionID;
  IpxSpxEntry();
}

word SPXInitialize(byte *major, byte *minor, word *maxconn, word *availconn)
{
  _BX = 0x10;
  _AX = 0x00;
  IpxSpxEntry();
  *major = _BH;
  *minor = _BL;
  *maxconn = _CX;
  *availconn = _DX;
  return _AL;
}

word SPXEstablishConnection(word *connectionID, TEventControlBlock far *ecb,
                            byte retrycount, byte watchdog)
{
  _BX = 0x11;
  _AL = retrycount;
  _AH = watchdog;
  _ES = FP_SEG(ecb);
  _SI = FP_OFF(ecb);
  IpxSpxEntry();
  *connectionID = _DX;
  return _AL;
}

word SPXGetConnectionStatus(word connectionID, TConnectionInfo far *conninfo)
{
  _BX = 0x15;
  _DX = connectionID;
  _ES = FP_SEG(conninfo);
  _SI = FP_OFF(conninfo);
  IpxSpxEntry();
  return _AL;
}

void SPXListenForConnection(TEventControlBlock far *ecb,
                            byte retrycount, byte watchdog)
{
  _BX = 0x12;
  _AL = retrycount;
  _AH = watchdog;
  _ES = FP_SEG(ecb);
  _SI = FP_OFF(ecb);
  IpxSpxEntry();
}

void SPXListenForSequencedPacket(TEventControlBlock far *ecb)
{
  _BX = 0x17;
  _ES = FP_SEG(ecb);
  _SI = FP_OFF(ecb);
  IpxSpxEntry();
}

void SPXSendSequencedPacket(word connectionID, TEventControlBlock far *ecb)
{
  _BX = 0x16;
  _DX = connectionID;
  _ES = FP_SEG(ecb);
  _SI = FP_OFF(ecb);
  IpxSpxEntry();
}

void SPXTerminateConnection(word connectionID, TEventControlBlock far *ecb)
{
  _BX = 0x13;
  _DX = connectionID;
  _ES = FP_SEG(ecb);
  _SI = FP_OFF(ecb);
  IpxSpxEntry();
}

#endif

//=============================================================================
// Utility functions
//
//=============================================================================

char *IpxGetErrorDesc(int errorcode)
{
  static char *text[] = {
    "Success",
    "Socket table is full",
    "Socket is already open",
    "Socket is not opened",
    "No path or destination",
    "Cannot cancel ECB",
    "ECB is in use",
    "ECB is not in use",
    "Data buffer is too long",
    "Already in use for either send or listen",
  };
  return (char*)text[errorcode];
}

word Swap(word from)
{
  word temp;
  swab((char*)&from,(char*)&temp,sizeof(word));
  return temp;
}

long Swap(long from)
{
  long temp;
  swab((char*)&from,(char*)&temp,sizeof(long));
  return temp;
}

void ZeroMemory(void *address, word length)
{
  memset(address,'\0',length);
}

/*#############################################################################
# End of file IPXSPX.CPP
#############################################################################*/

