/*#############################################################################
# UTILITY.H
#
# 30.09.1995. Codor
#
# Description
# -----------
# Master include file for BLiX C++ object oriented DOS library.
# Most classes are also implemented in the Windows library (cwinlib)
#############################################################################*/

#ifndef __UTILITY_H__
#define __UTILITY_H__

#include <dos.h>
#include <bios.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <time.h>

#include "types.h"

class TProfile;
class TDebug;
class TDebugMemory;

//=============================================================================
// Utility functions and structures
//
//=============================================================================

long hatoi(const char *s);
bool isfile(char *name);

//=============================================================================
// Roller functionality
// Various rollers controlled by interrupt handlers
//=============================================================================

enum rollerId {
  roUser1 = 0, roUser2, roUser3, roUser4, roUser5,
  roUser6, roUser7, roUser8, roUser9, roUser10,
};

void InitRoller(void);
void RefreshRoller(rollerId id);

//=============================================================================
// class TDebug
//
//=============================================================================

class TDebug {
  private:
    FILE *dbgFile;
    bool logTime;
  public:
    TDebug(const char *name, bool logtime);
    ~TDebug();
    void Log(char *text, ...);
    void Show(char *string, ...);
};

void TRACE(char *fmt, ...);

//=============================================================================
// class TPeriod
// Time interval measurement
//=============================================================================

class TPeriod {
  private:
    bool working;
    bool timeOut;
    long period;
    long start, stop;

  private:
    long getTickCount(long = 0L);

  public:
    TPeriod();
    TPeriod(long);
    TPeriod(word);
    void Start(word);
    void Start(long);
    bool Expired(void);
    bool Timeout(void);
    void Wait(void);
};

//=============================================================================
// class TDate, TTime
// These are date and time classes that replace class library TDate and TTime
// classes.
//=============================================================================

class TDate {
  private:
    unsigned char d, m;
    unsigned int y;
    char buff[15];
    void parse(const char *);

  public:
    enum prnFmt { LongYear, ShortYear, NoYear };

  public:
    TDate();
    TDate(const char *);
    TDate(unsigned char, unsigned char, unsigned int);
    void Refresh(void);
    char *AsString(prnFmt = ShortYear);
    void Assign(unsigned char, unsigned char, unsigned int);
    void Parse(const char *);
    void Set(void);
    unsigned char Day(void)   const { return d; }
    unsigned char Month(void) const { return m; }
    unsigned int  Year(void)  const { return y; }
    void operator=(const TDate&);
    int operator==(const TDate&) const;
};

class TTime {
  private:
    unsigned char h, m, s;
    char buff[10];
    void parse(const char *);

  public:
    enum prnFmt { Full, NoSec };

  public:
    TTime();
    TTime(const char *);
    TTime(unsigned char, unsigned char = 0, unsigned char = 0);
    void Refresh(void);
    char *AsString(prnFmt = Full);
    void Assign(unsigned char, unsigned char, unsigned char = 0);
    void Parse(const char *);
    void Set(void);
    unsigned char Hour(void) const  { return h; }
    unsigned char Min(void) const   { return m; }
    unsigned char Sec(void) const   { return s; }
    void operator=(const TTime &);
    int operator==(const TTime&) const;
};

//=============================================================================
// class TProfile
// This class describes the initialization file. This file and functions
// to access it are very similar to Windows INI file and associated
// functions
//=============================================================================

typedef void (*EnumerateSection)(char *, char *, char *, dword);

class TProfile {
  public:
    enum Error  { None = 0, NoIniFile, NoSection, NoEntry };
    enum Status { lineOk = 0, gotoNextLine, endOfFile };

    Error LastError;

  private:
    char  IniFile[81];
    int   Len;
    char *Entry, *Value, *Line, *Temp;
    FILE *iniFP;
    char  COMMENT, LGROUP, RGROUP;

  private:
    bool locateSection(char *section, word&);
    bool getEntry(word&, char *entry, char *value);
    bool isSection(char *section);
    bool isEntry(char *entry, char *value);
    Status  readIniLine(char *li);

  public:
    TProfile(char *, int = 1024);
    ~TProfile();
    bool  Open(void);
    void  Close(void);
    bool  isError(void) { return (LastError != None) ? True : False; }
    char *iniFile(void) { return (char *) &(IniFile[0]); }
    virtual bool Enum(char *section, EnumerateSection proc, dword data);
    virtual char *Get(char *section, char *entry, const char *deflt);
    virtual int Get(char *section, char *entry, int deflt);
    virtual long Get(char *section, char *entry, long deflt);
    virtual void Transient(void) {};
};

//=============================================================================
// class TDebugMemory
// The object of this class must be created first (and automatically destroyed
// last) so put it as a statis data member in module where main() is.
//=============================================================================

class TDebugMemory {
  private:
    ulong MemB,MemA;
  public:
    TDebugMemory();
    ~TDebugMemory();
};

#endif

/*#############################################################################
# End of UTILITY.H
#############################################################################*/

