/*#############################################################################
# VIDEO.CPP
#
# Autor: Todorovic Zoran
# Datum: Decembar '92.
#
# Funkcije koriste adresu na kojoj pocinje video memorija. Postoje dva seta
# funkcija i to funkcije koje ispisuju nesto na ekran i funkcije koje me-
# njaju atribut dela ekrana.
# Za normalan rad je potrebna adresa video memorije koja se nalazi u
# srukturi Config koja se puni funkcijom ScanHardware() u datoteci system.c
#
# Koordinate tacaka na ekranu su date na sledeci nacin:
#     col= (1-80)   row= (1-25)
#############################################################################*/

#include "video.h"

//
// Base video memory
//
static unsigned int VideoMemory;

#define MAX_COL           80
#define MAX_ROW           25
#define ISVALID(c,r)      (c >= 1 && c <= MAX_COL && r >= 1 && r <= MAX_ROW)
#define VIDEOBUF          (VideoMemory)
#define VIDEOOFS(c,r)     ((MAX_COL*(r-1) + (c-1)) << 1)

#define  ADDRESS(c,r) ((unsigned char far *) MK_FP(VIDEOBUF, VIDEOOFS(c,r)))

//
// Static (local) functions
//
static void format_border(char *, char *);
static char get_horizontal_cross_char(int, char []);
static char get_vertical_cross_char(int, char []);

/*#############################################################################
# INITIALIZATION
#
# Set the base address of video memory. If video card is not present then an
# argument could be introduced which is True if video card is absent. Then
# set the base address to virtual screen - allocated memory area of 4Kb.
#############################################################################*/

uint InitVideo(void)
{
  union REGS r;
  r.h.ah = 0x0F;
  int86(0x10, &r, &r);
  switch (r.h.al) {
    case 7:
      VideoMemory = 0xB000;
      break;
    case 2:
    case 3:
      VideoMemory = 0xB800;
      break;
  }
  return VideoMemory;
}

/*#############################################################################
#
#
#############################################################################*/

void CursorGoto(int col, int row)
{
  union REGS r;
  r.h.ah = 0x02;
  r.h.bh = 0;
  r.h.dh = row - 1;
  r.h.dl = col - 1;
  int86(0x10,&r,&r);
}

void CursorWhere(int *col, int *row)
{
  union REGS r;
  r.h.ah = 0x03;
  r.h.bh = 0;
  int86(0x10,&r,&r);
  *row = r.h.dh + 1;
  *col = r.h.dl + 1;
}

/*#############################################################################
# CHAR
#
# Funkcije za prikaz karaktera na odgovarajucu poziciju na ekranu.
#############################################################################*/

/******************************************************************************
* PlaceChar()
* Postavi karakter na lokaciju ekrana bez izmene atributa
******************************************************************************/

void PlaceChar(int col, int row, char ch)
{
  if (ISVALID(col,row))
    *ADDRESS(col,row) = ch;
}

/******************************************************************************
* PlaceCharAttr()
* Postavi karakter na lokaciju ekrana sa izmenom atributa
******************************************************************************/

void PlaceCharAttr(int col, int row, char attr, char ch)
{
  unsigned char far *videoptr;

  if (!ISVALID(col,row))
    return;

  videoptr = ADDRESS(col,row);
  *videoptr++ = ch;
  *videoptr   = attr;
}

/******************************************************************************
* PlaceAttr()
* Postavi atribut na jednu lokaciju ekrana
******************************************************************************/

void PlaceAttr(int col, int row, char attr)
{
  if (ISVALID(col,row))
    *(ADDRESS(col,row)+1) = attr;
}

/******************************************************************************
* RepeatChar()
* Postavi (num) karaktera na ekran pocevsi od lokacije (col,row) bez izmene
* attributa.
******************************************************************************/

void RepeatChar(int col, int row, int num, char ch)
{
  unsigned char far *videoptr;
  int i, endpos;

  if (!ISVALID(col,row))
    return;

  endpos = (col+num > 81) ? (81-col) : num;
  videoptr = ADDRESS(col,row);
  for (i= 0; i< endpos; i++) {
    *videoptr++ = ch;
    videoptr++;
  }
}

/******************************************************************************
* RepeatCharAttr()
* Postavi (num) karaktera na ekran pocevsi od lokacije (col,row) i to u speci-
* ficiranom attributu.
******************************************************************************/

void RepeatCharAttr(int col, int row, int num, char attr, char ch)
{
  unsigned char far *videoptr;
  int i, endpos;

  if (!ISVALID(col,row))
    return;

  endpos = (col+num > 81) ? (81-col) : num;
  videoptr = ADDRESS(col,row);
  for (i= 0; i< endpos; i++) {
    *videoptr++ = ch;
    *videoptr++ = attr;
  }
}

/*#############################################################################
# STRING
#
# Funkcije za prikaz stringa na odgovarajucu poziciju na ekranu.
#############################################################################*/

/******************************************************************************
* PutStrg()
* Prikazi string na lokaciji (col,row). Nemoj menjati atribut u kome se
* prikazuje string.
******************************************************************************/

void PutStrg(int col, int row, char *strg)
{
  unsigned char far *videoptr;
  int i, endpos, len= strlen(strg);

  if (!ISVALID(col,row))
    return;

  endpos = (col+len > 81) ? (81-col) : len;
  videoptr = ADDRESS(col,row);
  for (i= 0; i< endpos; i++) {
    *videoptr++ = strg[i];
    videoptr++;
  }
}

/******************************************************************************
* PutPartStrg()
* Prikazi deo stringa koji pocinje na ofsetu (pos) i duzine je (len) na
* ekransku lokaciju (col,row). Atribut u kome se prikazuje string se ne menja
******************************************************************************/

void PutPartStrg(int col, int row, int pos, int len, char *strg)
{
  unsigned char far *videoptr;
  int i, slen= strlen(strg);

  if (!ISVALID(col,row) || pos >= slen || len <= 0)
    return;

  if (pos+len > slen)
    len= slen-pos;

  len = (col+len > 81) ? (81-col) : len;
  videoptr = ADDRESS(col,row);
  for (i= 0; i< len; i++) {
    *videoptr++ = strg[pos+i];
    videoptr++;
  }
}

/******************************************************************************
* PutStrgTrail()
* Prikazi zadnjih (len) karaktera stringa (strg) na lokaciji (col,row).
******************************************************************************/

void PutStrgTrail(int col, int row, int len, char *strg)
{
  unsigned char far *videoptr;
  int i, slen= strlen(strg);

  if (!ISVALID(col,row))
    return;

  if (len > slen)
    len= slen;
  if (col+len > 81)
    len= 81-col;

  videoptr = ADDRESS(col,row);
  for (i= slen-len; i< slen; i++) {
    *videoptr++ = strg[i];
    videoptr++;
  }
}

/******************************************************************************
* PutStrgAttr()
* Prikazi string na lokaciji (col,row) i to u atributu (attr)
******************************************************************************/

void PutStrgAttr(int col, int row, char attr, char *strg)
{
  unsigned char far *videoptr;
  int i, endpos, len= strlen(strg);

  if (!ISVALID(col,row))
    return;

  endpos = (col+len > 81) ? (81-col) : len;
  videoptr = ADDRESS(col,row);
  for (i= 0; i< endpos; i++) {
    *videoptr++ = strg[i];
    *videoptr++ = attr;
  }
}

/******************************************************************************
* PutPartStrgAttr()
* Prikazi deo stringa koji pocinje na ofsetu (pos) i duzine je (len) na
* ekransku lokaciju (col,row). Za prikazivanje se koristi atribut (attr).
******************************************************************************/

void PutPartStrgAttr(int col, int row, int pos, int len, char attr, char *strg)
{
  unsigned char far *videoptr;
  int i, slen= strlen(strg);

  if (!ISVALID(col,row) || pos >= slen || len <= 0)
    return;

  if (pos+len > slen)
    len= slen-pos;

  len = (col+len > 81) ? (81-col) : len;
  videoptr = ADDRESS(col,row);
  for (i= 0; i< len; i++) {
    *videoptr++ = strg[pos+i];
    *videoptr++ = attr;
  }
}

/******************************************************************************
* PutStrgTrailAttr()
* Prikazi zadnjih (len) karaktera stringa (strg) na lokaciji (col,row).
******************************************************************************/

void PutStrgTrailAttr(int col, int row, int len, char attr, char *strg)
{
  unsigned char far *videoptr;
  int i, slen= strlen(strg);

  if (!ISVALID(col,row))
    return;

  if (len > slen)
    len= slen;
  if (col+len > 81)
    len= 81-col;

  videoptr = ADDRESS(col,row);
  for (i= slen-len; i< slen; i++) {
    *videoptr++ = strg[i];
    *videoptr++ = attr;
  }
}

/******************************************************************************
* PutAttr()
* Promeni atribut dela linije koji pocinje na lokaciji (col,row) i duzine je
* (len) u atribut (attr)
******************************************************************************/

void PutAttr(int col, int row, char len, char attr)
{
  unsigned char far *videoptr;
  int i, endpos;

  if (!ISVALID(col,row))
    return;

  endpos = (col+len > 81) ? (81-col) : len;
  videoptr = ADDRESS(col,row);
  for (i= 0; i< endpos; i++) {
    videoptr++;
    *videoptr++ = attr;
  }
}

/******************************************************************************
* CenterStrg()
* Centriraj string unutar duzine (width) pocevsi sa pozicijom (col,row) bez
* izmene atributa dela ekrana.
******************************************************************************/

void CenterStrg(int col, int row, int width, char *strg)
{
  unsigned char far *videoptr;
  int i, endpos, len= strlen(strg);

  if (! ISVALID(col,row) || len > width)
    return;

  col += (width-len) >> 1;
  endpos = (col+len > 81) ? (81-col) : len;
  videoptr = ADDRESS(col,row);
  for (i = 0; i < endpos; i++) {
    *videoptr++ = strg[i];
    videoptr++;
  }
}

/******************************************************************************
* CenterStrgAttr()
* Centriraj string unutar duzine (width) pocevsi sa pozicijom (col,row) sa
* izmenom atributa.
******************************************************************************/

void CenterStrgAttr(int col, int row, int width, char attr, char *strg)
{
  unsigned char far *videoptr;
  int i, endpos, len= strlen(strg);

  if (!ISVALID(col,row) || len > width)
    return;

  col += (width-len) >> 1;
  endpos = (col+len > 81) ? (81-col) : len;
  videoptr = ADDRESS(col,row);
  for (i= 0; i< endpos; i++) {
    *videoptr++ = strg[i];
    *videoptr++ = attr;
  }
}

/*#############################################################################
# BOX
#
# Funkcije za rad sa prozorom tj. delom ekrana.
#############################################################################*/

/******************************************************************************
* BoxFill()
* Punjenje box-a sa karakterom bez izmene atributa.
******************************************************************************/

void BoxFill(int col, int row, int width, int height, char ch)
{
  int i, j;
  unsigned char far *videoptr;

  if (!ISVALID(col,row))
    return;

  if (col+width > 81)
    width= 81-col;
  if (row+height > 26)
    height= 26-row;

  for (j= 0; j< height; j++) {
    videoptr = ADDRESS(col,row+j);
    for (i= 0; i< width; i++) {
      *videoptr++ = ch;
      videoptr++;
    }
  }
}

/******************************************************************************
* BoxFillAttr()
* Punjenje box-a sa karakterom sa izmenom atributa.
******************************************************************************/

void BoxFillAttr(int col, int row, int width, int height, char attr, char ch)
{
  int i, j;
  unsigned char far *videoptr;

  if (!ISVALID(col,row))
    return;

  if (col+width > 81)
    width= 81-col;
  if (row+height > 26)
    height= 26-row;

  for (j= 0; j< height; j++) {
    videoptr = ADDRESS(col,row+j);
    for (i= 0; i< width; i++) {
      *videoptr++ = ch;
      *videoptr++ = attr;
    }
  }
}

/******************************************************************************
* BoxAttr()
* Izmena atributa unutar box-a
******************************************************************************/

void BoxAttr(int col, int row, int width, int height, char attr)
{
  int i, j;
  unsigned char far *videoptr;

  if (!ISVALID(col,row))
    return;

  if (col+width > 81)
    width= 81-col;
  if (row+height > 26)
    height= 26-row;

  for (j= 0; j< height; j++) {
    videoptr = ADDRESS(col,row+j);
    for (i= 0; i< width; i++) {
      videoptr++;
      *videoptr++ = attr;
    }
  }
}

/*#############################################################################
# BORDER
#
# Funkcije za rad sa okvirima
#############################################################################*/

/******************************************************************************
* BorderChar()
* Nacrtaj okvir bez izmene atributa
******************************************************************************/

void BorderChar(int col, int row, int width, int height, char *borderstrg)
{
  int  i;
  char chars[8];

  if (!ISVALID(col,row))
    return;

  if (col+width > 81)
    width= 81-col;
  if (row+height > 26)
    height= 26-row;

  format_border(borderstrg, chars);

  PlaceChar(col, row, chars[0]);
  PlaceChar(col+width-1, row, chars[2]);
  PlaceChar(col, row+height-1, chars[6]);
  PlaceChar(col+width-1, row+height-1, chars[4]);

  RepeatChar(col+1, row, width-2, chars[1]);
  RepeatChar(col+1, row+height-1, width-2, chars[5]);
  for (i= 0; i< height-2; i++) {
    PlaceChar(col, row+1+i, chars[7]);
    PlaceChar(col+width-1, row+1+i, chars[3]);
  }
}

/******************************************************************************
* BorderCharAttr()
* Nacrtaj okvir sa izmenom atributa
******************************************************************************/

void BorderCharAttr(int col, int row, int width, int height,
                                               char attr, char *borderstrg)
{
  int  i;
  char chars[8];

  if (!ISVALID(col,row))
    return;

  if (col+width > 81)
    width= 81-col;
  if (row+height > 26)
    height= 26-row;

  format_border(borderstrg, chars);

  PlaceCharAttr(col, row, attr, chars[0]);
  PlaceCharAttr(col+width-1, row, attr, chars[2]);
  PlaceCharAttr(col, row+height-1, attr, chars[6]);
  PlaceCharAttr(col+width-1, row+height-1, attr, chars[4]);

  RepeatCharAttr(col+1, row, width-2, attr, chars[1]);
  RepeatCharAttr(col+1, row+height-1, width-2, attr, chars[5]);
  for (i= 0; i< height-2; i++) {
    PlaceCharAttr(col, row+1+i, attr, chars[7]);
    PlaceCharAttr(col+width-1, row+1+i, attr, chars[3]);
  }
}

/******************************************************************************
* BorderAttr()
* Izmeni atribut za specificirani okvir
******************************************************************************/

void BorderAttr(int col, int row, int width, int height, char attr)
{
  int  i;

  if (!ISVALID(col,row))
    return;

  if (col+width > 81)
    width= 81-col;
  if (row+height > 26)
    height= 26-row;

  PutAttr(col, row, width, attr);
  PutAttr(col, row+height-1, width, attr);
  for (i= 0; i< height-2; i++) {
    PlaceAttr(col, row+1+i, attr);
    PlaceAttr(col+width-1, row+1+i, attr);
  }
}

/******************************************************************************
* format_border()
* Formatiraj ivicu box-a
******************************************************************************/

static void format_border(char *borderstrg, char *chars)
{
  int   i, error= 0;
  char *valid= "ĳͺڿɻոַ";

  strupr(borderstrg);
  for (i = 0; i < strlen(borderstrg); i++)
    if (strchr("FSD", borderstrg[i]) == NULL) {
      error = 1;
      break;
    }
  if (i != 4)
    error = 1;

  if (error)
    strcpy(borderstrg, "SSSS");

  if (borderstrg[0] == 'F') {
    chars[0] = chars[6] = valid[20];
    chars[2] = chars[4] = valid[23];
    chars[1] = valid[22];
    chars[3] = valid[23];
    chars[5] = valid[21];
    chars[7] = valid[20];
  }
  else {
    /** gornja horizontalna linija **/
    if (borderstrg[0] == 'S') {
      chars[1]= valid[0];
      if (borderstrg[3] == 'S')      { chars[0]=valid[4];  chars[7]=valid[1]; }
      else if (borderstrg[3] == 'D') { chars[0]=valid[16]; chars[7]=valid[3]; }
      if (borderstrg[1] == 'S')      { chars[2]=valid[5];  chars[3]=valid[1]; }
      else if (borderstrg[1] == 'D') { chars[2]=valid[17]; chars[3]=valid[3]; }
    }
    else if (borderstrg[0] == 'D') {
      chars[1]= valid[2];
      if (borderstrg[3] == 'S')      { chars[0]=valid[12]; chars[7]=valid[1]; }
      else if (borderstrg[3] == 'D') { chars[0]=valid[8];  chars[7]=valid[3]; }
      if (borderstrg[1] == 'S')      { chars[2]=valid[13]; chars[3]=valid[1]; }
      else if (borderstrg[1] == 'D') { chars[2]=valid[9];  chars[3]=valid[3]; }
    }
    /** donja forizontalna linija **/
    if (borderstrg[2] == 'S') {
      chars[5]= valid[0];
      if (borderstrg[3] == 'S')      chars[6]= valid[7];
      else if (borderstrg[3] == 'D') chars[6]= valid[19];
      if (borderstrg[1] == 'S')      chars[4]= valid[6];
      else if (borderstrg[1] == 'D') chars[4]= valid[18];
    }
    else if (borderstrg[2] == 'D') {
      chars[5]= valid[2];
      if (borderstrg[3] == 'S')      chars[6]= valid[15];
      else if (borderstrg[3] == 'D') chars[6]= valid[11];
      if (borderstrg[1] == 'S')      chars[4]= valid[14];
      else if (borderstrg[1] == 'D') chars[4]= valid[10];
    }
  }
}

/*#############################################################################
# LINE
#
# Funkcije za crtanje linija. Ukrstanje sa vec postojecom linijom se resava
# sa odgovarajucim karakterima.
#############################################################################*/

/******************************************************************************
* HorizontalLine()
* Nacrtaj horizontalnu liniju sa korekcijom na ukrstanju sa postojecom linijom
******************************************************************************/

void HorizontalLine(int col, int row, int length, int linea)
{
  unsigned char far *videoptr;
  int   i, num;
  char  a[5];
  char  *up=  "¿˻Ѹҷ",
        *down="ٴʼϾн";

  if (!ISVALID(col,row))
    return;

  num = (col+length > 81) ? (81-col) : length;
  videoptr = ADDRESS(col,row);

  for (i = 0; i < num; i++) {
    a[0]= GetChar(col+i, row);
    a[1]= (i==0) ? 0 : GetChar(col+i-1, row);      /** prethodni **/
    a[2]= (i==num-1) ? 0 : GetChar(col+i+1, row);  /** sledeci   **/

    a[3]= (row==1) ? 0 : GetChar(col+i, row-1);    /** gornji    **/
    if (a[3] != 0 && strchr(up, a[3]) == NULL)     a[3]= 0;
    a[4]= (row==25) ? 0 : GetChar(col+i, row+1);   /** donji     **/
    if (a[4] != 0 && strchr(down, a[4]) == NULL)   a[4]= 0;

    *videoptr++ = get_horizontal_cross_char(linea, a);
    videoptr++;
  }
}

/******************************************************************************
* VerticalLine()
* Nacrtaj vertikalnu liniju sa korekcijom na ukrstanju sa postojecom linijom
******************************************************************************/

void VerticalLine(int col, int row, int length, int linea)
{
  unsigned char far *videoptr;
  int  i, num;
  char a[5];
  char  *left ="",
        *right="ٴ¼˾ѽ";

  if (!ISVALID(col,row))
    return;

  num = (row+length > 26) ? (81-row) : length;
  videoptr= ADDRESS(col,row);

  for (i = 0; i < num; i++) {
    a[0]= GetChar(col, row+i);
    a[1]= (i==0) ? 0 : GetChar(col, row+i-1);      /** prethodni **/
    a[2]= (i==num-1) ? 0 : GetChar(col, row+i+1);  /** sledeci   **/

    a[3]= (col==1) ? 0 : GetChar(col-1, row+i);    /** levi      **/
    if (a[3] != 0 && strchr(left, a[3]) == NULL)     a[3]= 0;
    a[4]= (col==80) ? 0 : GetChar(col+1, row+i);   /** desni     **/
    if (a[4] != 0 && strchr(right, a[4]) == NULL)    a[4]= 0;

    *videoptr= get_vertical_cross_char(linea, a);
    videoptr += 160;
  }
}

/******************************************************************************
* get_horizontal_cross_char()
* Vraca nazad karakter koji treba postaviti na mesto ukrstanja dve linije
******************************************************************************/

static char get_horizontal_cross_char(int linea, char a[5])
{
  #define THIS    a[0]
  #define W       a[1]        /** prethodni **/
  #define E       a[2]        /** sledeci **/
  #define N       a[3]        /** gornji **/
  #define S       a[4]        /** donji **/

  char retch;
  switch (linea) {
    case bidSingleBorder:
      if (THIS == '') {
        if      (W==0 && N==0)          retch= '';
        else if (W==0 && N!=0)          retch= '';
        else if (W==0 && S==0)          retch= '';
        else if (W==0 && S!=0)          retch= '';
        else if (E==0 && N==0)          retch= '';
        else if (E==0 && N!=0)          retch= '';
        else if (E==0 && S==0)          retch= '';
        else if (E==0 && S!=0)          retch= '';
        else if (W!=0 && E!=0 && N==0)  retch= '';
        else if (W!=0 && E!=0 && S==0)  retch= '';
        else                            retch= '';
      }
      else if (THIS == '') {
        if      (W==0 && N==0)          retch= '';
        else if (W==0 && N!=0)          retch= '';
        else if (W==0 && S==0)          retch= '';
        else if (W==0 && S!=0)          retch= '';
        else if (E==0 && N==0)          retch= '';
        else if (E==0 && N!=0)          retch= '';
        else if (E==0 && S==0)          retch= '';
        else if (E==0 && S!=0)          retch= '';
        else if (W!=0 && E!=0 && N==0)  retch= '';
        else if (W!=0 && E!=0 && S==0)  retch= '';
        else                            retch= '';
      }
      else retch= '';
      break;
    case bidDoubleBorder:
      if (THIS == '') {
        if      (W==0 && N==0)          retch= '';
        else if (W==0 && N!=0)          retch= '';
        else if (W==0 && S==0)          retch= '';
        else if (W==0 && S!=0)          retch= '';
        else if (E==0 && N==0)          retch= '';
        else if (E==0 && N!=0)          retch= '';
        else if (E==0 && S==0)          retch= '';
        else if (E==0 && S!=0)          retch= '';
        else if (W!=0 && E!=0 && N==0)  retch= '';
        else if (W!=0 && E!=0 && S==0)  retch= '';
        else                            retch= '';
      }
      else if (THIS == '') {
        if      (W==0 && N==0)          retch= '';
        else if (W==0 && N!=0)          retch= '';
        else if (W==0 && S==0)          retch= '';
        else if (W==0 && S!=0)          retch= '';
        else if (E==0 && N==0)          retch= '';
        else if (E==0 && N!=0)          retch= '';
        else if (E==0 && S==0)          retch= '';
        else if (E==0 && S!=0)          retch= '';
        else if (W!=0 && E!=0 && N==0)  retch= '';
        else if (W!=0 && E!=0 && S==0)  retch= '';
        else                            retch= '';
      }
      else retch= '';
      break;
  }
  return retch;

  #undef THIS
  #undef W
  #undef E
  #undef N
  #undef S
}

/******************************************************************************
* get_vertical_cross_char()
* Vraca nazad karakter koji treba postaviti na mesto ukrstanja dve linije
******************************************************************************/

static char get_vertical_cross_char(int linea, char a[5])
{
  #define THIS    a[0]
  #define N       a[1]            /** prethodni **/
  #define S       a[2]            /** sledeci **/
  #define W       a[3]            /** levi **/
  #define E       a[4]            /** desni **/

  char retch;
  switch (linea) {
    case bidSingleBorder:
      if (THIS == '') {
        if      (W==0 && N==0)          retch= '';
        else if (W==0 && N!=0)          retch= '';
        else if (W==0 && S==0)          retch= '';
        else if (W==0 && S!=0)          retch= '';
        else if (E==0 && N==0)          retch= '';
        else if (E==0 && N!=0)          retch= '';
        else if (E==0 && S==0)          retch= '';
        else if (E==0 && S!=0)          retch= '';
        else if (W!=0 && E!=0 && N==0)  retch= '';
        else if (W!=0 && E!=0 && S==0)  retch= '';
        else                            retch= '';
      }
      else if (THIS == '') {
        if      (W==0 && N==0)          retch= '';
        else if (W==0 && N!=0)          retch= '';
        else if (W==0 && S==0)          retch= '';
        else if (W==0 && S!=0)          retch= '';
        else if (E==0 && N==0)          retch= '';
        else if (E==0 && N!=0)          retch= '';
        else if (E==0 && S==0)          retch= '';
        else if (E==0 && S!=0)          retch= '';
        else if (W!=0 && E!=0 && N==0)  retch= '';
        else if (W!=0 && E!=0 && S==0)  retch= '';
        else                            retch= '';
      }
      else retch= '';
      break;
    case bidDoubleBorder:
      if (THIS == '') {
        if      (W==0 && N==0)          retch= '';
        else if (W==0 && N!=0)          retch= '';
        else if (W==0 && S==0)          retch= '';
        else if (W==0 && S!=0)          retch= '';
        else if (E==0 && N==0)          retch= '';
        else if (E==0 && N!=0)          retch= '';
        else if (E==0 && S==0)          retch= '';
        else if (E==0 && S!=0)          retch= '';
        else if (W!=0 && E!=0 && N==0)  retch= '';
        else if (W!=0 && E!=0 && S==0)  retch= '';
        else                            retch= '';
      }
      else if (THIS == '') {
        if      (W==0 && N==0)          retch= '';
        else if (W==0 && N!=0)          retch= '';
        else if (W==0 && S==0)          retch= '';
        else if (W==0 && S!=0)          retch= '';
        else if (E==0 && N==0)          retch= '';
        else if (E==0 && N!=0)          retch= '';
        else if (E==0 && S==0)          retch= '';
        else if (E==0 && S!=0)          retch= '';
        else if (W!=0 && E!=0 && N==0)  retch= '';
        else if (W!=0 && E!=0 && S==0)  retch= '';
        else                            retch= '';
      }
      else retch= '';
      break;
  }
  return retch;

  #undef THIS
  #undef N
  #undef S
  #undef W
  #undef E
}

/*#############################################################################
# READ CHAR & ATTR
#
# Funkcije za citanje karaktera/atributa direktno sa ekrana
#############################################################################*/

/******************************************************************************
* GetChar()
* Procitaj karaker sa pozicije na ekranu
******************************************************************************/

char GetChar(int col, int row)
{
  if (ISVALID(col,row))
    return *ADDRESS(col,row);
  return 0;
}

/******************************************************************************
* GetAttr()
* Procitaj karaker sa pozicije na ekranu
******************************************************************************/

char GetAttr(int col, int row)
{
  if (ISVALID(col,row))
    return *(ADDRESS(col,row)+1);
  return 0;
}

/******************************************************************************
* GetCharAttr()
* Procitaj karaker sa pozicije na ekranu
******************************************************************************/

void GetCharAttr(int col, int row, char *ch, char *attr)
{
  if (ISVALID(col,row)) {
    *ch = *(ADDRESS(col,row));
    *attr = *(ADDRESS(col,row)+1);
  }
}

/*#############################################################################
# READ & WRITE BUFFER
#
# Smesti deo ekrana u bafer ili bafer na ekran
#############################################################################*/

/******************************************************************************
* GetCharArea()
* Procitaj deo ekrana u karakter bafer
******************************************************************************/

void GetCharArea(int col, int row, int width, int height, char *buffer)
{
  int i, j;
  unsigned char far *videoptr;

  if (!ISVALID(col,row)) {
    buffer= NULL;
    return;
  }

  for (j= 0; j< height; j++) {
    videoptr = ADDRESS(col,row+j);
    for (i= 0; i< width; i++) {
      *buffer++= *videoptr++;
      videoptr++;
    }
  }
}

/******************************************************************************
* GetCharAttrArea()
* Procitaj deo ekrana u bafer. Cita se karakter i atribut
******************************************************************************/

void GetCharAttrArea(int col, int row, int width, int height, char *buffer)
{
  int i, j;
  unsigned char far *videoptr;

  if (!ISVALID(col,row)) {
    buffer= NULL;
    return;
  }

  for (j= 0; j< height; j++) {
    videoptr = ADDRESS(col,row+j);
    for (i= 0; i< width; i++) {
      *buffer++= *videoptr++;
      *buffer++= *videoptr++;
    }
  }
}

/******************************************************************************
* PutCharArea()
* Prebaci sadrzaj bafera u video memoriju
******************************************************************************/

void PutCharArea(int col, int row, int width, int height, char *buffer)
{
  int i, j;
  unsigned char far *videoptr;

  if (!ISVALID(col,row)) {
    buffer= NULL;
    return;
  }

  for (j= 0; j< height; j++) {
    videoptr = ADDRESS(col,row+j);
    for (i= 0; i< width; i++) {
      *videoptr++= *buffer++;
      videoptr++;
    }
  }
}

/******************************************************************************
* PutCharAttrArea()
* Smesti bafer (karakter i atribut) u video memoriju
******************************************************************************/

void PutCharAttrArea(int col, int row, int width, int height, char *buffer)
{
  int i, j;
  unsigned char far *videoptr;

  if (!ISVALID(col,row)) {
    buffer= NULL;
    return;
  }

  for (j= 0; j< height; j++) {
    videoptr = ADDRESS(col,row+j);
    for (i= 0; i< width; i++) {
      *videoptr++= *buffer++;
      *videoptr++= *buffer++;
    }
  }
}

/*#############################################################################
# SCROLL
#
# Funkcije za skrolovanje dela ekrana na gore ili na dole
#############################################################################*/

/******************************************************************************
* ScrollUp()
* Funkcija skroluje deo ekrana na gore i to za odredjeni broj linija
******************************************************************************/

void ScrollUp(int col, int row, int width, int height, int num_lines)
{
  int i, j;
  unsigned char far *vptr_src;
  unsigned char far *vptr_dst;

  if (!ISVALID(col,row))
    return;

  if (col + width > MAX_COL + 1)
    width = MAX_COL + 1 - col;
  if (row + height > MAX_ROW + 1)
    height = MAX_ROW + 1 - row;

  for (i = num_lines, j= 0; i < height; i++, j++) {
    vptr_src = ADDRESS(col,row+i);
    vptr_dst = ADDRESS(col,row+j);
    memmove(vptr_dst, vptr_src, width << 1);
  }
  for (i = height-num_lines; i < height; i++) {
    vptr_src = ADDRESS(col,row+i);
    for (j = 0; j < width; j++) {
      *vptr_src++ = ' ';
      vptr_src++;
    }
  }
}

/******************************************************************************
* ScrollDown()
* Funkcija skroluje deo ekrana nadole i to za odredjeni broj linija
******************************************************************************/

void ScrollDown(int col, int row, int width, int height, int num_lines)
{
  int i, j;
  unsigned char far *vptr_src;
  unsigned char far *vptr_dst;

  if (!ISVALID(col,row))
    return;

  if (col + width > MAX_COL + 1)
    width = MAX_COL + 1 - col;
  if (row + height > MAX_ROW + 1)
    height = MAX_ROW + 1 - row;

  for (i = height-num_lines-1, j = height-1; i >= 0; i--, j--) {
    vptr_src = ADDRESS(col,row+i);
    vptr_dst = ADDRESS(col,row+j);
    memmove(vptr_dst, vptr_src, width << 1);
  }
  for (i = 0; i < num_lines; i++) {
    vptr_src = ADDRESS(col,row+i);
    for (j = 0; j < width; j++) {
      *vptr_src++ = ' ';
      vptr_src++;
    }
  }
}

/*#############################################################################
# Kraj datoteke VIDEO.C
#############################################################################*/

